from typing import Union
import warnings


class OrkgResponse(object):

    @property
    def succeeded(self) -> bool:
        return str(self.status_code)[0] == '2'

    def __init__(self, response, status_code: str, content: Union[list, dict], url: str, paged: bool):
        if response is None and status_code is None and content is None and url is None:
            raise ValueError("either response should be provided or content with status code")
        if not paged:
            warnings.warn("You are running an out-of-date ORKG backend! GET (i.e., listing) calls now provide pagination information", DeprecationWarning)
        if (response is not None and str(response.status_code)[0] != '2') or (status_code is not None and content is not None and status_code[0] != '2'):
            self.status_code = response.status_code if response is not None else status_code
            self.content = response.content if response is not None else content
            self.pageable = None
            return
        if response is not None:
            self.status_code = response.status_code
            self.content = response.json() if len(response.content) > 0 else response.content
            self.pageable = None
            if paged and isinstance(self.content, dict):
                self.pageable = self.content['pageable'] if 'pageable' in self.content else None
                self.content = self.content['content'] if 'content' in self.content else self.content
            self.url = response.url
        if status_code is not None and content is not None:
            self.status_code = status_code
            self.content = content
            self.pageable = None
            if paged and isinstance(self.content, dict) and 'content' in content:
                self.pageable = self.content['pageable'] if 'pageable' in self.content else None
                self.content = self.content['content']
            self.url = url

    def __repr__(self):
        return "%s %s" % ("(Success)" if self.succeeded else "(Fail)", self.content)

    def __str__(self):
        return self.content



from __future__ import unicode_literals

from datetime import date, datetime
from functools import wraps
from urllib.parse import quote_plus, urlencode
import warnings

# parts of URL to be omitted
SKIP_IN_PATH = (None, "", b"", [], ())


def _escape(value):
    """
    Escape a single value of a URL string or a query parameter. If it is a list
    or tuple, turn it into a comma-separated string first.
    """

    # make sequences into comma-separated stings
    # if isinstance(value, (list, tuple)):
    #     value = ",".join(value)

    # dates and datetimes into isoformat
    if isinstance(value, (date, datetime)):
        value = value.isoformat()

    # make bools into true/false strings
    elif isinstance(value, bool):
        value = str(value).lower()

    # don't decode bytestrings
    elif isinstance(value, bytes):
        return value

    # encode strings to utf-8
    # if isinstance(value, string_types):
    #     return value.encode("utf-8")

    if isinstance(value, (list, tuple)):
        return value
    return str(value)


def _make_path(*parts):
    """
    Create a URL string from parts, omit all `None` values and empty strings.
    Convert lists and tuples to comma separated values.
    """
    return "/" + "/".join(
        # preserve ',' and '*' in url for nicer URLs in logs
        quote_plus(_escape(p), b",*")
        for p in parts
        if p not in SKIP_IN_PATH
    )


# parameters that apply to all methods
GLOBAL_PARAMS = ("pretty", "human", "error_trace", "format", "filter_path")


def query_params(*query_params):
    """
    Decorator that pops all accepted parameters from method's kwargs and puts
    them in the params argument.
    """

    def _wrapper(func):
        @wraps(func)
        def _wrapped(*args, **kwargs):
            params = {}
            if "params" in kwargs:
                params = kwargs.pop("params").copy()
            for p in query_params + GLOBAL_PARAMS:
                if p in kwargs:
                    v = kwargs.pop(p)
                    if v is not None:
                        params[p] = _escape(v)

            # don't treat ignore and request_timeout as other params to avoid escaping
            for p in ("ignore", "request_timeout"):
                if p in kwargs:
                    params[p] = kwargs.pop(p)
            return func(*args, params=params, **kwargs)

        return _wrapped

    return _wrapper


def dict_to_url_params(params):
    return "?%s" % urlencode(params)


def simcomp_available(func):
    def check_if_simcomp_available(self, *args, **kwargs):
        if not self.client.simcomp_available:
            raise ValueError("simcomp_host must be provided in the ORKG wrapper class!")
        return func(self, *args, **kwargs)

    return check_if_simcomp_available


def admin_functionality(func):
    def display_admin_warning(self, *args, **kwargs):
        # TODO: create some functionality in the backend to check for permissions
        warnings.warn("This call needs elevated role in the system!", RuntimeWarning)
        return func(self, *args, **kwargs)

    return display_admin_warning


class NamespacedClient(object):
    def __init__(self, client):
        self.client = client
        if self.client.token is not None:
            self.auth = {'Authorization': f'Bearer {self.client.token}'}
        else:
            self.auth = None

    @staticmethod
    def handle_sort_params(params: dict):
        if 'sort' in params:
            sort_direction = 'asc'
            if 'desc' in params:
                sort_direction = 'desc' if params['desc'] else 'asc'
                del params['desc']
            params['sort'] = f"{params['sort']},{sort_direction}"



from orkg.utils import NamespacedClient
from orkg.out import OrkgResponse
from typing import Dict, List, Union


class DummyClient(NamespacedClient):

    def create_xxx_response(self, code: str, content: Union[Dict, List]) -> OrkgResponse:
        return OrkgResponse(status_code=code, content=content, paged=False, response=None, url="")

    def create_200_response(self, content: Union[Dict, List]) -> OrkgResponse:
        return self.create_xxx_response("200", content)

    def create_404_response(self, content: Union[Dict, List]) -> OrkgResponse:
        return self.create_xxx_response("404", content)



from unittest import TestCase
from orkg import ORKG


class TestTemplates(TestCase):
    """
    Some test scenarios might need to be adjusted to the content of the running ORKG instance
    """
    orkg = ORKG(host="https://sandbox.orkg.org")

    def test_materialize(self):
        self.orkg.templates.materialize_templates(templates=['R199091', 'R199040'], verbose=False)
        self.assertTrue(True)

    def test_df_template(self):
        from pandas import DataFrame as df
        lst = ['this', 'is', 'fancy']
        lst2 = [4, 2, 5]
        param = df(list(zip(lst, lst2)), columns=['word', 'length'])
        self.orkg.templates.materialize_template(template_id='R199091')
        self.orkg.templates.test_df(label="what!", dataset=(param, 'Fancy Table'), uses_library="pyORKG").pretty_print(format='json-ld')
        self.assertTrue(True)




from orkg.utils import NamespacedClient, simcomp_available
from orkg.out import OrkgResponse
from typing import List, Tuple
import pandas as pd
from ast import literal_eval
from urllib.parse import urlparse, parse_qs


class ContributionsClient(NamespacedClient):

    @simcomp_available
    def similar(self, cont_id: str) -> OrkgResponse:
        self.client.simcomp._append_slash = True
        response = self.client.simcomp.similar(cont_id).GET()
        return self.client.wrap_response(response=response)

    @simcomp_available
    def compare(self, contributions: List[str], response_hash: str = None) -> OrkgResponse:
        self.client.simcomp._append_slash = False
        params = f'?contributions={",".join(contributions)}'
        if response_hash is not None:
            params = f'{params}&response_hash={response_hash}'
        response = self.client.simcomp.compare.GET(params)
        return self.client.wrap_response(response=response)

    def __get_contributions_from_comparison(self, comparison_id: str) -> Tuple[List[str], str, List[str]]:
        resource = self.client.resources.by_id(comparison_id).content
        if 'Comparison' not in resource['classes']:
            raise ValueError("This id is not for a comparison")
        response = self.client.json.get_json(resource_id=comparison_id).content
        query_params = parse_qs(urlparse(response['data']['url']).query)
        return query_params['contributions'][0], query_params['response_hash'][0], query_params['properties'][0].split(',') if 'properties' in query_params else []

    @simcomp_available
    def compare_dataframe(self, contributions: List[str] = None, comparison_id: str = None, like_ui = True) -> pd.DataFrame:
        ui_props = []
        if contributions is None and comparison_id is None:
            raise ValueError("either provide the contributions, or the comparison ID")
        response_hash = None
        if comparison_id is not None:
            contributions, response_hash, ui_props = self.__get_contributions_from_comparison(comparison_id)
        response = self.compare(contributions=contributions, response_hash=response_hash)
        if not response.succeeded:
            return pd.DataFrame()
        content = response.content
        contributions_list = content['contributions']
        columns = [f"{contribution['title']}/{contribution['contributionLabel']}" for contribution in
                   contributions_list]
        properties_list = content['properties']
        property_lookup = {prop['id']: prop['label'] for prop in properties_list}
        # create table view of the data
        data = content['data']
        indices = []
        rows = []
        for prop_id, values in data.items():
            indices.append(property_lookup[prop_id])
            row = []
            for index, value in enumerate(values):
                if not value[0]:
                    row.append('')
                else:
                    if len(value) == 1:
                        try:
                            # Try parse the single value into datatype
                            row.append(literal_eval(value[0]['label']))
                        except (SyntaxError, ValueError):
                            # If it fails, that means it is a string
                            row.append(value[0]['label'])
                    else:
                        cell = []
                        # Rather than parsing it in a list comprehension that might fail
                        # Then the whole list will be typed as strings, we parse it one item at a time
                        for v in value:
                            try:
                                # Try parse each value by itself
                                cell.append(literal_eval(v['label']))
                            except (SyntaxError, ValueError):
                                # If it fails, that means it is a string
                                cell.append(v['label'])
                        row.append(cell)
            rows.append(row)
        # create data frame from pieces
        df = pd.DataFrame(rows, columns=columns, index=indices)
        if like_ui and len(ui_props) > 0:
            # remove props that are not displayed
            df = df.drop([p_lbl for p_id, p_lbl in property_lookup.items() if p_id not in ui_props])
            # order df rows by ui props order
            df = df.reindex([property_lookup[p] for p in ui_props if p in property_lookup])
        return df



from orkg.utils import NamespacedClient, query_params, dict_to_url_params
from orkg.out import OrkgResponse


class StatementsClient(NamespacedClient):

    def by_id(self, id) -> OrkgResponse:
        self.client.backend._append_slash = True
        response = self.client.backend.statements(id).GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get(self, params=None) -> OrkgResponse:
        self.handle_sort_params(params)
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.statements.GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.statements.GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_by_subject(self, subject_id, params=None) -> OrkgResponse:
        self.handle_sort_params(params)
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.statements.subject(subject_id).GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.statements.subject(subject_id).GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_by_predicate(self, predicate_id, params=None) -> OrkgResponse:
        self.handle_sort_params(params)
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.statements.predicate(predicate_id).GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.statements.predicate(predicate_id).GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_by_object(self, object_id, params=None) -> OrkgResponse:
        self.handle_sort_params(params)
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.statements.object(object_id).GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.statements.object(object_id).GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_by_object_and_predicate(self, object_id, predicate_id, params=None) -> OrkgResponse:
        self.handle_sort_params(params)
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.statements.object(object_id).predicate(predicate_id).GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.statements.object(object_id).predicate(predicate_id).GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_by_subject_and_predicate(self, subject_id, predicate_id, params=None) -> OrkgResponse:
        self.handle_sort_params(params)
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.statements.subject(subject_id).predicate(predicate_id).GET(
                dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.statements.subject(subject_id).predicate(predicate_id).GET()
        return self.client.wrap_response(response)

    @query_params("subject_id", "predicate_id", "object_id")
    def add(self, params=None) -> OrkgResponse:
        if len(params) != 3:
            raise ValueError("all parameters must be provided")
        else:
            self.client.backend._append_slash = True
            params['object'] = {'id': params['object_id'], '_class': 'literal' if params['object_id'][0] == 'L' else 'resource'}
            response = self.client.backend.statements.POST(json=params, headers=self.auth)
        return self.client.wrap_response(response)

    @query_params("subject_id", "predicate_id", "object_id")
    def update(self, id, params=None) -> OrkgResponse:
        if len(params) == 0:
            raise ValueError("at least one parameter must be provided")
        else:
            if not self.exists(id):
                raise ValueError("the provided id is not in the graph")
            self.client.backend._append_slash = True
            response = self.client.backend.statements(id).PUT(json=params, headers=self.auth)
        return self.client.wrap_response(response)

    @query_params("minLevel", "maxLevel", "blacklist", "whitelist", "includeFirst")
    def bundle(self, thing_id, params=None) -> OrkgResponse:
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.statements(thing_id).bundle.GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.statements(thing_id).bundle.GET()

        return self.client.wrap_response(response)

    def delete(self, id) -> OrkgResponse:
        if not self.exists(id):
            raise ValueError("the provided id is not in the graph")
        self.client.backend._append_slash = True
        response = self.client.backend.statements(id).DELETE()
        return self.client.wrap_response(response)

    def exists(self, id) -> bool:
        return self.by_id(id).succeeded



import networkx as nx

from typing import List, Union
from requests.exceptions import RequestException

from orkg import ORKG


def subgraph(client: ORKG, thing_id: str, blacklist: Union[str, List[str]] = '', max_level: int = -1) -> nx.DiGraph:
    """
    Obtains a networkx directed graph representation of any ORKG component given by its thing_id.
    E.g. of ORKG components: Paper, Contribution, Comparison, Template

    It starts from the thing_id resource and traverses the graph until all literals are reached.

    :param client: orkg.ORKG client used to connect with ORKG backend.
    :param thing_id: Any subject, object or predicate ID in the ORKG.
    :param blacklist: Class(es) to be excluded from the subgraph. E.g. 'ResearchField'
        (see `orkgc:ResearchField <https://orkg.org/class/ResearchField>`_).
        Note that the first subgraph level will always be included.
    :param max_level: Deepest subgraph's level to traverse.
    """
    blacklist = [blacklist] if isinstance(blacklist, str) else blacklist

    try:
        response = client.statements.bundle(thing_id, params={
            'includeFirst': 'true',
            'blacklist': ','.join(blacklist),
            'maxLevel': max_level
        })
    except RequestException:
        response = None

    if not response or not response.succeeded:
        raise ValueError('Something went wrong while connecting to ORKG backend with host {}'.format(client.host))

    statements = response.content['statements']

    if not statements:
        raise ValueError('Nothing found for the provided ID: {}'.format(thing_id))

    return _construct_subgraph(statements)


def _construct_subgraph(statements: list) -> nx.DiGraph:
    """
    Constructs a subgraph represented by the given RDF ``statements``.

    :param statements: List of all RDF statements describing the subgraph.
    """
    _subgraph = nx.DiGraph()

    for statement in statements:
        # NetworkX does not create a node or an edge double, therefore, the following implementation works :)
        start_node = _create_node_from_thing(_subgraph, statement['subject'])
        target_node = _create_node_from_thing(_subgraph, statement['object'])
        _create_edge(_subgraph, start_node, target_node, statement['predicate'])

    return _subgraph


def _create_node_from_thing(_subgraph: nx.DiGraph, thing: dict) -> str:
    _subgraph.add_node(node_for_adding=thing['id'], **thing)
    return thing['id']


def _create_edge(_subgraph: nx.DiGraph, start_node: str, target_node: str, predicate: dict) -> None:
    _subgraph.add_edge(start_node, target_node, **predicate)



from unittest import TestCase
from orkg import ORKG
from tempfile import NamedTemporaryFile


class TestResources(TestCase):
    """
    Some test scenarios might need to be adjusted to the content of the running ORKG instance
    """
    orkg = ORKG()

    def test_by_id(self):
        res = self.orkg.resources.by_id('R1')
        self.assertTrue(res.succeeded)

    def test_get(self):
        size = 10
        res = self.orkg.resources.get(size=size)
        self.assertTrue(res.succeeded)
        self.assertEqual(len(res.content), size)

    def test_get_with_items(self):
        count = 30
        res = self.orkg.resources.get(size=count)
        self.assertTrue(res.succeeded)
        self.assertEqual(len(res.content), count)

    def test_delete(self):
        to_delete = self.orkg.resources.add(label="I should be deleted")
        print(to_delete.content)
        self.assertTrue(to_delete.succeeded)
        res = self.orkg.resources.delete(id=to_delete.content["id"])
        self.assertTrue(res.succeeded)
        self.assertFalse(self.orkg.resources.exists(to_delete.content["id"]))

    def test_add(self):
        label = "test"
        res = self.orkg.resources.add(label=label)
        self.assertTrue(res.succeeded)
        self.assertEqual(res.content['label'], label)

    def test_find_or_add(self):
        import random
        import string
        label = ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(15))
        old = self.orkg.resources.add(label=label)
        self.assertTrue(old.succeeded, 'Creating first resource is a success')
        self.assertEqual(old.content['label'], label, 'The first resource has the correct label')
        new = self.orkg.resources.find_or_add(label=label)
        self.assertTrue(new.succeeded, 'Creating second resource is a success')
        self.assertEqual(new.content['id'], old.content['id'], 'the two resources have the same id')

    def test_add_with_class(self):
        label = "test"
        cls = "Coco"
        res = self.orkg.resources.add(label=label, classes=[cls])
        self.assertTrue(res.succeeded)
        self.assertEqual(res.content['label'], label)
        self.assertIn(cls, res.content['classes'])

    def test_update(self):
        res = self.orkg.resources.add(label="Coco")
        self.assertTrue(res.succeeded)
        label = "test"
        res = self.orkg.resources.update(id=res.content['id'], label=label)
        self.assertTrue(res.succeeded)
        res = self.orkg.resources.by_id(res.content['id'])
        self.assertTrue(res.succeeded)
        self.assertEqual(res.content['label'], label)

    def test_add_tabular_data(self):
        csv_content = """Statistics,Event start time,Event end time,Duration,Sunrise,Sunset
Min,09:05,14:25,04:41,04:34,16:29
Max,12:17,18:41,06:32,07:50,19:53
Mean,10:58,16:46,05:47,06:13,18:12
Median,11:07,16:51,05:43,06:15,18:14
"""
        label = "test dataset XXI"

        csv_file = NamedTemporaryFile(mode="w", delete=False)
        csv_file.write(csv_content)
        csv_path = csv_file.name
        csv_file.close()

        res = self.orkg.resources.save_dataset(
            csv_path, label,
            ["Statistics", "Event start time", "Event end time", "Duration", "Sunrise", "Sunset"]
        )

        self.assertTrue(res.succeeded)
        self.assertEqual(res.content['label'], label)



from orkg.utils import NamespacedClient
from orkg.out import OrkgResponse


class StatsClient(NamespacedClient):

    def get(self) -> OrkgResponse:
        self.client.backend._append_slash = True
        response = self.client.backend.stats.GET()
        return self.client.wrap_response(response)



from unittest import TestCase
from orkg import ORKG
from tempfile import NamedTemporaryFile


class TestObjects(TestCase):
    """
    Some test scenarios might need to be adjusted to the content of the running ORKG instance
    """
    orkg = ORKG()

    def test_add(self):
        obj = {
            "predicates": [],
            "resource": {
                "name": "I rock maybe!",
                "classes": [
                    "C2000"
                ],
                "values": {
                    "P32": [
                        {
                            "@id": "R2"
                        }
                    ],
                    "P55": [
                        {
                            "label": "ORKG is so cool!",
                            "classes": [
                                "C3000"
                            ]
                        }
                    ]
                }
            }
        }
        res = self.orkg.objects.add(params=obj)
        self.assertTrue(res.succeeded)



from unittest import TestCase
from orkg import ORKG


class TestBackend(TestCase):
    """
    Some test scenarios might need to be adjusted to the content of the running ORKG instance
    """

    def test_auth(self):
        orkg = ORKG(host='http://127.0.0.1:8080', creds=('test@test.test', 'test123'))
        self.assertIsNotNone(orkg.token)





from orkg.utils import NamespacedClient
from orkg.out import OrkgResponse
from orkg.convert import JsonLDConverter
from typing import Dict, List, Union, Tuple, Optional
import re
from tqdm import tqdm
import json
import types
import keyword
import datetime
import pandas as pd

NATIVE_TYPES = ['Number', 'String', 'Boolean', 'Integer', 'Date', 'URI']
DATAFRAME_CLASSES = ['QBDataset', 'Table']


def pre_process_string(string: str) -> str:
    lower_string = string.lower().replace(' ', '_')
    res = re.sub(r'\W+', '', lower_string)
    return res


def check_against_keywords(string: str) -> str:
    """
    Checks a string against python keywords.
    If a keyword is found, it is postfixed with a '_template'.
    """
    if string in keyword.kwlist:
        return string + '_template'
    return string


def clean_up_dict(dictionary: dict) -> dict:
    """
    replaces the key 'name' with 'label' inside the 'resource' object.
    then strips the root key 'resource' and returns the inner dictionary.
    """
    dictionary = dictionary['resource']
    dictionary['label'] = dictionary['name']
    del dictionary['name']
    return dictionary


def remove_empty_values(dictionary: dict):
    if 'values' not in dictionary['resource']:
        return
    for _, value in dictionary['resource']['values'].items():
        for v in value:
            if not bool(v):
                value.remove(v)


def map_type_to_pythonic_type(value_type: str) -> str:
    if value_type == 'Number':
        return 'Union[int, float, complex]'
    elif value_type == 'Boolean':
        return 'bool'
    elif value_type == 'Integer':
        return 'int'
    elif value_type == 'Date':
        return 'datetime.date'
    else:
        return 'str'


class ObjectStatement(object):
    # Optionals
    label: Optional[str]
    classes: Optional[List[str]]
    text: Optional[str]
    datatype: Optional[str]
    values: Optional[List]
    name: Optional[str]
    # Must have
    predicate_id: str
    template_id: str

    def __init__(self, predicate_id: str, template_id: str):
        self.predicate_id = predicate_id
        self.template_id = template_id
        self.label = None
        self.classes = None
        self.text = None
        self.datatype = None
        self.values = None
        self.name = None

    @staticmethod
    def create_main(name: str, classes: List[str] = None) -> 'ObjectStatement':
        statement = ObjectStatement('', '')
        statement.set_main_statement(name, classes)
        return statement

    @property
    def is_main_statement(self) -> bool:
        return self.name is not None

    def set_main_statement(self, name: str, classes: List[str] = None):
        self.name = name
        self.classes = classes

    def set_literal(self, text: str, datatype: Optional[str] = None):
        self.text = text
        self.datatype = datatype

    def set_nested_statement(self, statement: 'ObjectStatement'):
        self.set_resource(statement.label, statement.classes)
        for value in statement.values:
            self.add_value(value)

    def set_resource(self, label: str, classes: List[str] = None):
        self.label = label
        self.classes = classes

    def add_value(self, value: 'ObjectStatement'):
        if self.values is None:
            self.values = []
        self.values.append(value)

    def values_to_statements(self) -> str:
        if self.values is None:
            return '{}'
        return '{' + ', '.join([f'"{v.predicate_id}": [{{{v.to_statement()}}}]' for v in self.values]) + '}'

    def to_statement(self) -> str:
        statement = ""
        if self.text is not None:
            statement += f'"text": {self.text}'
            if self.datatype is not None:
                statement += f', "datatype": "{self.datatype}"'
        elif self.label is not None:
            statement += f'"label": {self.label}'
            if self.classes is not None:
                statement += f', "classes": {self.classes}'
            if self.values is not None:
                statement += f', "values": {self.values_to_statements()} '
        return statement

    def serialize(self) -> str:
        if self.is_main_statement:
            return f'{{ "resource": {{ "name": "{self.name}", "classes": {self.classes}, "values": {self.values_to_statements()} }} }}'
        else:
            return f'"{self.predicate_id}": [{{{self.to_statement()}}}]'


class TemplateInstance(object):
    """
    A class that takes a dictionary in the constructor
    With methods to serialize the dictionary to a file, pretty print it, and send it to the KG.
    """

    def __init__(self, template_dict, orkg_client) -> None:
        self.template_dict = template_dict
        self.client = orkg_client
        self.preprocess_dict()

    def preprocess_dict(self):
        if isinstance(self.template_dict, str):
            self.template_dict = self.template_dict.strip("'<>() ").replace('\'', '\"')
            self.template_dict = json.loads(self.template_dict)

    def serialize_to_file(self, file_path: str, format: str = "orkg") -> None:
        """
        Serializes the template to a file.
        :param format: the format of the serialization (default: "orkg", possible: "json-ld")
        :param file_path: the file path to save the template to
        """
        with open(file_path, 'w') as f:
            if format.lower().strip() == "json-ld":
                json.dump(JsonLDConverter(self.client.host).convert_2_jsonld(self.template_dict), f, indent=4)
            else:
                json.dump(self.template_dict, f, indent=4)

    def pretty_print(self, format: str = "orkg") -> None:
        """
        Pretty prints the template to the console.
        :param format: the format of the printed text (default: "orkg", possible: "json-ld")
        """
        if format.lower().strip() == "json-ld":
            print(json.dumps(JsonLDConverter(self.client.host).convert_2_jsonld(self.template_dict), indent=4))
        else:
            print(json.dumps(self.template_dict, indent=4))

    def save(self) -> OrkgResponse:
        """
        Saves the template to the server.
        :return: The OrkgResponse from the server
        """
        return self.client.objects.add(params=self.template_dict)

    @staticmethod
    def parse_dict(dictionary: Dict, obj: ObjectStatement) -> None:
        if 'resource' in dictionary:
            # This is a complete resource and not a nested one
            dictionary = dictionary['resource']
            obj.set_resource(dictionary['name'], dictionary['classes'])
            if 'values' in dictionary:
                dictionary = dictionary['values']
        for key, values in dictionary.items():
            for value in values:
                sub_obj = ObjectStatement(key, obj.template_id)
                obj.add_value(sub_obj)
                if 'text' in value:
                    datatype = value.get('datatype', None)
                    sub_obj.set_literal(f"\"{value['text']}\"", datatype)
                elif 'label' in value:
                    classes = value.get('classes', None)
                    sub_obj.set_resource(f"\"{value['label']}\"", classes)
                    if 'values' in value:
                        TemplateInstance.parse_dict(value['values'], sub_obj)


class TemplateComponent:
    predicate_id: str
    predicate_label: str
    value_class_id: Optional[str]
    value_class_label: Optional[str]
    is_of_custom_type: bool
    min_cardinality: Optional[int]
    max_cardinality: Optional[int]
    is_nested_template: bool
    nested_template_target: Optional[str]

    # private properties
    _orkg_client: NamespacedClient

    def __init__(self, orkg_client, component_id: str):
        self._orkg_client = orkg_client
        component_statements = orkg_client.statements.get_by_subject(subject_id=component_id, size=99999).content
        # Set predicate info
        predicate = list(
            filter(lambda x: x['predicate']['id'] == 'TemplateComponentProperty', component_statements)
        )[0]['object']
        self.predicate_id = predicate['id']
        self.predicate_label = predicate['label']
        # Set class value, id, and label
        value_class = list(
            filter(lambda x: x['predicate']['id'] == 'TemplateComponentValue', component_statements)
        )
        if len(value_class) == 0:
            self.value_class_id = 'String'
            self.value_class_label = 'Text'
            self.is_of_custom_type = False
        else:
            value_class = value_class[0]['object']
            self.value_class_id = value_class['id']
            self.value_class_label = value_class['label']
            self.is_of_custom_type = True if self.value_class_id not in NATIVE_TYPES else False
        # Set cardinality of the property
        max_cardinality = list(
            filter(lambda x: x['predicate']['id'] == 'TemplateComponentOccurrenceMax', component_statements)
        )
        self.max_cardinality = None if len(max_cardinality) == 0 else max_cardinality[0]['object']['label']
        mix_cardinality = list(
            filter(lambda x: x['predicate']['id'] == 'TemplateComponentOccurrenceMin', component_statements)
        )
        self.mix_cardinality = None if len(mix_cardinality) == 0 else mix_cardinality[0]['object']['label']
        # Set if the component is nested template
        if not self.is_of_custom_type:
            self.is_nested_template = False
        else:
            template_classes = orkg_client.statements.get_by_object_and_predicate(object_id=self.value_class_id,
                                                                                  predicate_id='TemplateOfClass',
                                                                                  size=99999).content
            if len(template_classes) == 0:
                self.is_nested_template = False
            else:
                self.is_nested_template = True
                self.nested_template_target = template_classes[0]['subject']['id']
                # Recursively materialize nested templates of the root
                self._orkg_client.templates.materialize_template(self.nested_template_target)

    def get_clean_name(self) -> str:
        return pre_process_string(check_against_keywords(self.predicate_label))

    def get_return_type(self) -> str:
        # Check if the return type should be a pandas.DataFrame
        if self.value_class_id in DATAFRAME_CLASSES:
            return 'Union[pd.DataFrame, Tuple[pd.DataFrame, str]]'
        # Check if it is a nested template
        if self.is_nested_template:
            return 'TemplateInstance'
        # o/w it is a pure python type
        return map_type_to_pythonic_type(self.value_class_id)

    def get_property_as_function_parameter(self) -> str:
        clean_label = self.get_clean_name()
        return_type = self.get_return_type()
        return f'{clean_label}: {return_type}'

    def get_param_documentation(self) -> str:
        if self.is_nested_template:
            target_template = Template.find_or_create_template(self._orkg_client, self.nested_template_target)
            return f':param {self.get_clean_name()}: a nested template, use orkg.templates.{target_template.get_template_name_as_function_name()}'
        else:
            # component is not nested template, return normal documentation
            return f':param {self.get_clean_name()}: a parameter of type {self.value_class_label}'

    def create_object_statement(self) -> ObjectStatement:
        to_str = self.get_return_type() == 'datetime.date'
        value = self.get_clean_name()
        statement = ObjectStatement(self.predicate_id, "UNKNOWN")  # FIXME: template_id is unknown in this scope
        if not self.is_nested_template:
            if self.is_of_custom_type:
                statement.set_resource(
                    label=f'str({value})' if to_str else value,
                    classes=[self.value_class_id]
                )
            else:
                statement.set_literal(
                    text=f'str({value})' if to_str else value,
                    datatype=None  # FIXME: datatype is not set for literals
                )
        return statement

    def __str__(self):
        return f'{"N" if self.is_nested_template else ""}Property(id="{self.predicate_id}", label="{self.predicate_label}", class="{self.value_class_id}", cardinality=({self.min_cardinality}, {self.max_cardinality}))'


class Template(object):
    # Static variable to keep track of all templates
    templates: Dict = {}

    # Instance variables
    template_class: str
    is_formatted: bool
    template_id: str
    template_name: str
    components: List[TemplateComponent]

    def __init__(self, orkg_client, template_id: str):
        # Fetch template statements
        self.components = []
        template_statements = orkg_client.statements.get_by_subject(subject_id=template_id, size=99999).content
        # Iterate over template statements and create template components
        components = filter(lambda x: x['predicate']['id'] == 'TemplateComponent', template_statements)
        for component in components:
            self.components.append(TemplateComponent(orkg_client, component['object']['id']))
        # Set template class
        self.template_class = list(
            filter(lambda x: x['predicate']['id'] == 'TemplateOfClass', template_statements)
        )[0]['object']['id']
        # Set template info
        self.template_id = template_id
        self.template_name = orkg_client.resources.by_id(id=template_id).content['label']
        self.is_formatted = len(
            list(filter(lambda x: x['predicate']['id'] == 'TemplateLabelFormat', template_statements))
        ) > 0
        # Register template to the global templates dict
        Template.templates[template_id] = self

    @staticmethod
    def find_or_create_template(orkg_client, template_id: str) -> 'Template':
        """
        Check if template is in registry, if yes return it.
        Otherwise, instantiate a new template.
        """
        if template_id in Template.templates:
            return Template.templates[template_id]
        else:
            return Template(orkg_client, template_id)

    def get_params_as_function_params(self) -> Tuple[str, str]:
        """
        Returns a tuple of the function parameters for the template.
        The first element is the function parameters as a string,
        the second element is the function docstring as a string.
        """
        params = ', '.join([c.get_property_as_function_parameter() for c in self.components])
        if not self.is_formatted:
            params = f'label: str, {params}'
            if params.strip()[-1] == ',':
                params = params.strip()[:-1]
        params_docstring = '\n\t'.join(
            [
                comp.get_param_documentation()
                for comp in self.components
            ]
        )
        if not self.is_formatted:
            params_docstring = f':param label: the label of the resource of type string\n\t{params_docstring}'
        return params, params_docstring

    def get_template_name_as_function_name(self) -> str:
        return check_against_keywords(pre_process_string(self.template_name))

    def create_api_object_values_and_classes(self) -> Tuple[str, str]:
        """
        Returns the API object values for the template.
        """
        properties = [c.create_object_statement() for c in self.components if c.value_class_id not in DATAFRAME_CLASSES]
        return f""""values": {{{','.join([p.serialize() for p in properties])}}}""", f'"classes": ["{self.template_class}"]'

    def create_api_object(self):
        """
        Returns the complete ORKG API object for the template as string.
        """
        values, classes = self.create_api_object_values_and_classes()
        object_json = f"""{{
                "resource": {{
                    "name": {'""' if self.is_formatted else 'label'},
                    {classes},
                    {values}
                }}
            }}"""
        return object_json

    def __str__(self):
        return f'Template(id="{self.template_id}", class="{self.template_class}", components= "{len(self.components)} comps."))'


class OTFFunctionWriter(object):

    @staticmethod
    def convert_df_to_list_of_lists(df: pd.DataFrame, label: str = None):
        result = [[c for c in df.columns]]
        for index, row in df.iterrows():
            result.append([row[c] for c in df.columns])
        to_return = {'@df': result}
        if label is not None:
            to_return['label'] = label
        return to_return

    @staticmethod
    def implement_function(
            orkg_context: NamespacedClient,
            template_id: str
    ):
        template = Template.find_or_create_template(orkg_context.client, template_id)
        params, params_docstring = template.get_params_as_function_params()
        object_json = template.create_api_object()
        lookup_map = {component.get_clean_name(): component.predicate_id for component in template.components}
        function_name = check_against_keywords(pre_process_string(template.template_name))
        new_method = f'''
def {function_name}(self, {params}) -> TemplateInstance:
    """ Creates a template of type {template_id} ({template.template_name})

    {params_docstring}
    :return: a string representing the resource ID of the newly created resource
    """
    lookup_map = {lookup_map}
    obj = ObjectStatement.create_main({'""' if template.is_formatted else 'label'}, ['{template.template_class}'])
    TemplateInstance.parse_dict({object_json},obj)
    obj = TemplateInstance(obj.serialize(), self.client).template_dict
    for param_name, nested_template in {{k: v for k, v in locals().items() if isinstance(v, TemplateInstance) or isinstance(v, pd.DataFrame) or isinstance(v, Tuple)}}.items():
        predicate_id = lookup_map[param_name]
        if 'values' not in obj['resource']:
            obj['resource']['values'] = {{}}
        if not predicate_id in obj['resource']['values']:
            obj['resource']['values'][predicate_id] = []
        if isinstance(nested_template, pd.DataFrame):
            obj['resource']['values'][predicate_id].append(OTFFunctionWriter.convert_df_to_list_of_lists(nested_template))
        elif isinstance(nested_template, Tuple):
            obj['resource']['values'][predicate_id].append(OTFFunctionWriter.convert_df_to_list_of_lists(*nested_template))
        else:
            obj['resource']['values'][predicate_id].append(clean_up_dict(nested_template.template_dict))
    remove_empty_values(obj)
    return TemplateInstance(obj, self.client)

orkg_context.client.templates.{function_name} = types.MethodType( {function_name}, orkg_context )
orkg_context.client.templates.materialized_templates.append('{function_name}')
                    '''
        exec(new_method)


class TemplatesClient(NamespacedClient):
    materialized_templates: List[str] = []

    def materialize_template(self, template_id: str):
        """
        Materialize a singular ORKG template as python function
        :param template_id: the template id
        :return: True if everything is OK
        """
        return self._fetch_and_build_template_function(template_id)

    def materialize_templates(self, templates: Optional[List[str]] = None, verbose: bool = True):
        """
        Materialize a list of templates (or all templates if not provided)
        :param templates: optional list of templates
        :param verbose: sets if the process shows a progress bar or suppresses it
        """
        if templates is None:
            templates = [template['id'] for template in
                         self.client.classes.get_resource_by_class(class_id='ContributionTemplate', size=1000).content]
        exclude = ['R12000', 'R38504', 'R108555', 'R111221', 'R111231', 'R70785', 'R48000', 'R111155']  # RF, RP, Comp last 3
        iterator = tqdm(templates, desc='Materializing templates') if verbose else templates
        for template_id in iterator:
            if template_id not in exclude:
                self.materialize_template(template_id)

    def _fetch_and_build_template_function(self, template_id: str):
        """
        Internal function to create python digital twins of ORKG templates
        :param template_id: template ID to build
        :return: True if everything is OK
        """
        OTFFunctionWriter.implement_function(self, template_id)
        return True

    def list_templates(self):
        """
        List the set of materialized template functions
        :return: list of strings
        """
        return self.materialized_templates

    def get_template_specifications(self, template_id: str) -> str:
        """
        Return JSON representation of a template's specification (schema)
        :param template_id: the template to lookup
        :return: string representation of a JSON object
        """
        template = Template(self.client, template_id)
        result = {comp.get_clean_name(): f'A value of type ({comp.value_class_label})' for comp in template.components}
        return json.dumps(result, indent=4, sort_keys=True, default=str)

    def create_template_instance(self, template_id: str, instance: Dict) -> str:
        """
        Creates an instance of a given template by filling in
        the specifications of the template with the provided values

        :param template_id: the string representing the template ID
        :param instance: the json object that contains the components of the template instance
        """
        template = Template(self.client, template_id)
        obj = template.create_api_object()
        for key, value in instance.items():
            obj = obj.replace(key, f'"{value}"')
        json_object = json.loads(obj)
        return self.client.objects.add(params=json_object).content['id']



from unittest import TestCase

import networkx as nx

import orkg


class TestGraph(TestCase):
    """
    Some test scenarios might need to be adjusted to the content of the running ORKG instance
    """
    client = orkg.ORKG(host='https://incubating.orkg.org/')

    def test_subgraph(self):
        statements, contribution, adj = self._create_contribution()

        subgraph = orkg.subgraph(self.client, contribution)
        self.assertIsInstance(subgraph, nx.DiGraph)

        self.assertEqual(4, len(subgraph.nodes))
        self.assertEqual(3, len(subgraph.edges))

        for statement in statements:
            start_node = subgraph.nodes[statement.content['subject']['id']]
            start_label = statement.content['subject']['label']
            self.assertEqual(start_node['label'], start_label)

            target_node = subgraph.nodes[statement.content['object']['id']]
            target_label = statement.content['object']['label']
            self.assertEqual(target_node['label'], target_label)

            edge = subgraph.edges[start_node['id'], target_node['id']]
            edge_label = statement.content['predicate']['label']
            self.assertEqual(edge['label'], edge_label)

        self._delete_subgraph(statements)

    def _delete_subgraph(self, statements):
        for statement in statements:
            res = self.client.statements.delete(id=statement.content['id'])
            self.assertTrue(res.succeeded)

    def _create_contribution(self):
        """
        Contribution -> predicate_1 -> label_1;
                        predicate_2 -> object_1 .
        object_1 -> predicate_3 -> label_2
        """
        contribution = self.client.resources.add(label="Contribution").content['id']
        predicate_1 = self.client.predicates.add(label="predicate_1").content['id']
        label_1 = self.client.literals.add(label="label_1").content['id']
        statement_1 = self.client.statements.add(subject_id=contribution, predicate_id=predicate_1, object_id=label_1)

        predicate_2 = self.client.predicates.add(label="predicate_2").content['id']
        object_1 = self.client.resources.add(label="object_1").content['id']
        statement_2 = self.client.statements.add(subject_id=contribution, predicate_id=predicate_2, object_id=object_1)

        predicate_3 = self.client.predicates.add(label="predicate_3").content['id']
        label_2 = self.client.literals.add(label="label_2").content['id']
        statement_3 = self.client.statements.add(subject_id=object_1, predicate_id=predicate_3, object_id=label_2)

        adj = {contribution: {label_1: predicate_1, object_1: predicate_2}, object_1: {label_2: predicate_3}}
        return [statement_1, statement_2, statement_3], contribution, adj


